/*             CSE 687 Final Project --- Pointer Class 

            Student Name: Gang Cheng SSN:223-61-3010               */

/*******************************************************************
 * final.cc ---- main program to test the member & frind functions *
 *               for class Matrix and class Pointer                *
 *******************************************************************/

#include <fstream.h>  // cin, cout, cerr
#include "matrix.hh"  // class Matrix
#include "pointer.hh" // class Pointer

//--- <void error(char *s)> ---//
/* error handler */
void error(char *s)
{
  cerr << s; 
  exit(1);
} 

//--- < main program > ---//
/* test the member and friend functions for class Matrix
   which are rewritten using class Pointer and class Pointer itself */
main(int argc, char *argv[])
{
  char fname1[10], fname2[10];

  switch (argc) {
  case 1:
    cerr << "Input file name for matrix 1: ";
    cin >> fname1;
    cerr << "Input file name for matrix 2: ";
    cin >> fname2;
    break;
  case 2:
    strcpy(fname1,argv[1]);
    cerr << "Input file name for matrix 2: ";
    cin >> fname2;
    break;
  case 3:
    strcpy(fname1,argv[1]);
    strcpy(fname2,argv[2]);
    break;
  default:
    error("error: incorrect command line input\n");
    break;
  };

  // open matrix input files
  ifstream fp1(fname1);
  ifstream fp2(fname2);
 
  // check i/o status
  if (!fp1) {
    cerr << "cannot open file: " << fname1 << "\n";
    exit(1);
  };
  if (!fp2) {
    cerr << "cannot open file: " << fname2 << "\n";
    exit(1);
  };

  // status information     
  cerr << "\nThese are files " << fname1 << " and " << fname2 
       << "\nProcessed by " <<  argv[0] << "\n\n";

  // Below are sample testing statments designed specifically
  //  for testing some functionlities of Pointer class which 
  //  are not found in the testing program for Matrix Class.
  //
  // The member function tested for Matrix Class are:
  //   1. void constructor Matrix();
  //   2. constructor Matrix(int);
  //   3. (default) destructor ~Matrix();
  //   4. operator = ;
  //   5. operator + ;
  //   6. operator * ;
  //   7. istream operator >> ;
  //   8. ostream operator << ;
  //   9. member function matrix_size();
  //   
  // The member functions tested for Pointer Class are:
  //   1. void constructor Pointer() ;
  //   2. promoted copy constructor Pointer(T*) ;
  //   3. (default) destructor ~Pointer() ;
  //   4. memory allocator newptr(int) ;
  //   5. memory deallocator deleteptr() ;
  //   6. operator * ;
  //   7. operator [] ;
  //   8. operator + ;
  //   9. operator ++ (prefix form) ;
  //  10. operator ++ (postfix form) ;
  //  11. operator -- (prefix form) ;
  //  12. operator=(Pointer&) ;
  //  
  // Both pointers to auto variable and to allocated memory
  //  are used in a mixed way to illustrate the nontrival
  //  functionality of Pointer class

  // define 3 pointers to matrix
  Pointer<Matrix> mptr,mp1, mp2;

  // mptr points to an array(2 elements) of matrix,
  // each matrix has size 0x0 at start
  mptr.newptr(2);

  // mp1, mp2 each points to a matrix of size 0x0 
  mp1.newptr(); mp2.newptr(); 

  // read 2 matrixes and make mptr pointing to them, 
  // the size and actual memory allocated for each matrix
  //  is determined from the input files 
  fp1 >> *(mptr.operator++(1)); // unfortunately, g++ on Cat does not
  fp2 >> *mptr;                 //  accept the postfix form invocation
                                //  of ++(or --), I have to use this 
                                //  awkward form instead
  // make mp1 and mp2 pointing to the same matrixes as mptr does
  mp1 = --mptr; 
  mp2 = mptr + 1;

  // define a new matrix (auto variable) with the same size as 2nd matrix 
  Matrix m(mptr->matrix_size());

  // m = matrix1 + matrix2
  m = mptr[0] + mptr[1];

  // (incorrectly) try to delete the memory mptr pointing to
  mptr.deleteptr();
  
  // output results
  // because mp1 and mp2 point to the same memory as mptr does,
  //  it illustrates the (incorrect) deletion of mptr's memory
  //  did not make trouble to other pointers to the same one
  cout << "\nm1 = \n"<<*mp1 << "\nm2 = \n" <<*mp2;
  cout << "\nm1 + m2 =  \n" << m;

  // following segements illustrate the mixing use of pointers
  //  pointing to allocated memory and to auto variables
  Pointer<Matrix> mp3(&m);
  *mp3 = (*mp1)*(*mp2);
  cout << "\nm1 * m2 =  \n" << m;

  Pointer<Matrix> mp4 = mp3;
  *mp4 = *mp1 + *mp4;  
  cout << "\nm1 + m1 * m2 =  \n" << *mp4;

  // more than once deletions of same memory is save now
  mp1.deleteptr(); // first deletion
  mp2.deleteptr();

  mp1.deleteptr(); // second deletion
  mp2.deleteptr();

  return 0;
}
